"""
Convert pd.DataFrame with coordinates to KML file

Created on 12-05-2013. Migrated to Acequia on 27-06-2020
Depends on package SimpleKML

"""

import os
import os.path
import warnings
from collections import OrderedDict

import pandas as pd
from pandas import DataFrame, Series
import numpy as np
#import simplekml as skml
from acequia.geo.coordinate_conversion import CrdCon as CrdCon
import acequia as aq

sep = ','


class WpKml:
    """ Save Pandas DataFrame with waypoints in RD-coordinates to KML """

    iconshapes = {'square' : 'http://maps.google.com/mapfiles/kml/shapes/placemark_square.png',
                  'circle' : 'http://maps.google.com/mapfiles/kml/shapes/placemark_circle.png',
                  'cross'  : 'http://maps.google.com/mapfiles/kml/shapes/cross-hairs.png'}

    colors = {
      'white'       : '#FFFFFF',
      'brightgreen' : '#33FF00',
      'mossgreen'   : '#CCFF00',
      'deepyellow'  : '#FFFF00',
      'darkyellow'  : '#FFCC00',
      'orange'      : '#FF6600',
      'lightblue'   : '#99CCFF',
      'deepblue'    : '#0000FF',
      'brightpurple': '#CC00CC',
      'darkpurple'  : '#6600CC',
      'pink'        : '#FF00FF',
      'brightred'   : '#FF0000'}


    _default_styledict = OrderedDict([
        ('1',{'iconshape':'square','iconscale':1.0,'iconcolor':'#0c7cac',
              'labelscale':0.7,'labelcolor':'#FFFFFF'}),
        ('2',{'iconshape':'square','iconscale':1.0,'iconcolor':'#ac0b35',
              'labelscale':0.7,'labelcolor':'#FFFFFF'}),
        ('3',{'iconshape':'square','iconscale':1.0,'iconcolor':'#9e00b5',
              'labelscale':0.7,'labelcolor':'#FFFFFF'}),
        ('4',{'iconshape':'square','iconscale':1.0,'iconcolor':'#54a910',
              'labelscale':0.7,'labelcolor':'#FFFFFF'}),
        ('5',{'iconshape':'square','iconscale':1.0,'iconcolor':'#f69601',
              'labelscale':0.7,'labelcolor':'#FFFFFF'}),
        ('6',{'iconshape':'square','iconscale':1.0,'iconcolor':'#0011ff',
              'labelscale':0.7,'labelcolor':'#FFFFFF'}),
        ('7',{'iconshape':'square','iconscale':1.0,'iconcolor':'#ff00c1',
              'labelscale':0.7,'labelcolor':'#FFFFFF'}),
        ('9',{'iconshape':'square','iconscale':1.0,'iconcolor':'#ffff00',
              'labelscale':0.0,'labelcolor':'#FFFFFF'}),
        ])


    missing_style = { 'iconshape':'cross',
                      'iconscale':1.0,
                      'iconcolor':'#FFFFFF',
                      'labelscale':0.7,
                      'labelcolor':'#FFFFFF'}


    def __repr__(self):
        return 'Save Waypoints to KML'


    def __init__(self,wplist,label=None,xcoor=None,ycoor=None,
        styledict=None, stylecol=None):
        """Create KML file object from pd.DataFrame with waypoints


        Parameters
        ----------
        wplist : pd.DataFrame
            points with coordinates given in Standard Dutch Grid
        label : str
            columnname in wplist used for naming kml waypoints
        xcoor : float, optinal
            xcoordinate in RD
        ycoor : float, optional
            ycoordinate in RD
        styledic : dictionary, optional
            style definitions for each marker
        stylecol : str, optional
            column name for assigning styles


        Examples
        --------
        >>>sourcepath = '.\\testdata\\waypoints\\pmgloc.csv'
        >>>pmgloc = pd.read_csv(sourcepath)
        >>>pts = WpKml(pmgloc,label="nitgcode",xcoor="xcr",ycoor="ycr")
        >>>outdir = '.\\output\\kml\\'
        >>>pts.writekml(f'{outdir}pmv2019-noformatting.kml')

        >>>styledict_truefalse = {
                "true":
                    {"iconshape":"circle","iconscale":1.0,
                     "iconcolor":"#0000FF","labelscale":0.7,
                     "labelcolor":"FFFFFF"}),
                {"false":
                    {"iconshape":"circle","iconscale":1.0,
                     "iconcolor":"#FF00FF","labelscale":0.7,
                     "labelcolor":"FFFFFF"},}
        >>>pts = WpKml(pmgloc,label="nitgcode",xcoor="xcr",ycoor="ycr",
                       styledict=styledict_truefalse, stylecol="ispmg")
        >>>pts.writekml(f'{outdir}pmv-styledict-and-stylecol.kml')


        Notes
        -----
        If no values for xcoor,ycoor are given, column names with
        coordinates will be inferred from wplist column names.
        
        """

        self.kml = skml.Kml()
        self.defaultstyle = self._basestyle()

        if not isinstance(wplist, pd.DataFrame):
            raise TypeError('Variable wplist must be a pandas dataframe.')        

        if wplist.empty:
            raise ValueError('Variable wplist cannot be empty.')

        self.wplist = wplist.copy()
        self.wplist.columns = self.wplist.columns.str.lower()

        # set coordinates
        self.xcoor,self.ycoor = self._coordinate_columns(xcoor,ycoor)

        notlon = 'lon' not in self.wplist.keys()
        notlat = 'lat' not in self.wplist.keys()
        if notlon or notlat:
            self.wplist['lon'],self.wplist['lat'] = self.wgs84_coords()

        # set label column name
        if label is None:
            self.label = self.wplist.keys()[0]

        if label.lower() in self.wplist.keys():
            self.label = label.lower()
        else:
            firstcol = self.wplist.keys()[0]
            self.label = firstcol
            msg = ''.join(
                  f'Value {label} for label column not found in ',  
                  f'{self.wplist} column names. Column {firstcol} is ',
                  f'used for KML labels.')
            warnings.warn(msg,UserWarning)

        if stylecol is not None:
            stylecol = stylecol.lower()

        # set kml waypoint styles
        # TODO: Test function _validate_styledict
        self.styledict, self.stylecol = self._validate_styledict(styledict,stylecol)
        self.pointstyles = OrderedDict()
        for iconkey,icondef in self.styledict.items():
            basestyle = self._basestyle()
            self.pointstyles[iconkey] = self._changestyle(basestyle,icondef)


    def _coordinate_columns(self,xcoor=None,ycoor=None):
        """Return names of columns with xcoor,ycoor """

        if xcoor is None:
            xlist = ['xcoor','xrd','x']
            for x in xlist:
                if x in self.wplist.keys(): 
                    xcol = x
        else:
            if xcoor.lower() in self.wplist.keys(): 
                xcol = xcoor.lower()
            else:
                msg = f'No column with xcoor value available.'
                raise ValueError(msg)

        if ycoor is None:
            ylist = ['ycoor','yrd','y']
            for y in ylist:
                if y in self.wplist.keys(): 
                    yxol = y
        else:
            if ycoor.lower() in self.wplist.keys():
                ycol = ycoor.lower()
            else:
                msg = f'No column with ycoor value available.'
                raise ValueError(msg)

        return xcol,ycol


    def wgs84_coords(self):
        """Return lon and lat as ps.Series

        Returns
        -------
        lon,lat
        """
        crdcon = CrdCon()
        lon = self.wplist.apply(
                lambda x: crdcon.RDtoWGS84(float(x[self.xcoor]),
                float(x[self.ycoor]))['Lon'],axis=1)
        lat = self.wplist.apply(
                lambda x: crdcon.RDtoWGS84(float(x[self.xcoor]),
                float(x[self.ycoor]))['Lat'],axis=1)
        return lon,lat


    def default_styledict(self):
        """Return default styledict """
        return self._default_styledict


    def _validate_styledict(self,styledict,stylecol):
        """ validate styledict and stylecol """

        if (stylecol is not None) and (styledict is not None):

            # fill nans in column stylecol
            nanrows = self.wplist[self.wplist[stylecol].isnull()]
            if len(nanrows) !=0:
                fillnans = self.wplist[stylecol].replace(
                            np.nan,'missing',inplace=False)
                self.wplist[stylecol] = fillnans
                styledict['missing'] = self.missing_style


            # check styledict for missing keys
            # change stylecol to None if keys are missing from styledict
            missing = list(set(self.wplist[stylecol].unique()) 
                           - set(styledict.keys()))
            if len(missing)!=0: #not all(elem in wplabels for elem in styledict.keys()):               
                msg = ''.join([
                        f'Custom style dictionary styledict is missing ',
                        f'labels from waypoint column {stylecol}. ',
                        f'No formatting applied.\\n',
                        f'Missing labels are: {missing}'])
                warnings.warn(msg,UserWarning)
                stylecol = None


        if (stylecol is not None) and (styledict is None):

            wplabels = self.wplist[stylecol].unique()
            defaultkeys = list(self._default_styledict.keys())
            if len(wplabels) <= len(defaultkeys):
                newstyles = []
                for i,wplabel in enumerate(wplabels):
                    newstyles.append((wplabel,self._default_styledict[defaultkeys[i]]))
                styledict = OrderedDict(newstyles)
                #self.pointstyles = self._styles_from_dict(self.styledict)

            else:
                lenwp = len(wplabels)
                lenkeys = len(defaultkeys)
                message = ''.join(
                  [f'\nTo many unique values in waypoint stylecolumn '
                   f'{stylecol}.\n',
                   f'Stylecolumn {stylecol} contains {lenwp} values.\n',
                   f'Default style definitions contains {lenkeys} values.\n',
                   f'No formatting of waypoints applied.\n',])
                warnings.warn(message,UserWarning)
                styledict = {'default' : self._default_styledict['1']}
                stylecol = None


        if (stylecol is None) and (styledict is not None):
            styledict = next(iter(styledict.values())) # abitrary single element from styledict

        if (stylecol is None) and (styledict is None):
                styledict = {'default' : self._default_styledict['1']}


        """
        elif ((self.stylecol is None) and isinstance(self.styledict,dict)):
        if (stylecol is None) and (styledict is not None):
        if isinstance(styledict,dict):
            self.pointstyles = self._styles_from_dict(styledict)
        """
        return styledict,stylecol


    def _styles_from_dict(self,icondict):
        """Return OrderedDict of KML pointstyles with keys from icondict
        as point style names 

        Parameters
        ----------
        icondict : dict
            valid iconstyle dictionary


        Returns
        -------
        Ordereddict with KML pointstyles

        """
        pointstyles = OrderedDict()
        for iconkey,icondef in icondict.items():
            pointstyles[iconkey] = self._changestyle(
                                   kmlstyle=self._basestyle(),
                                   icondef=icondef)
        return pointstyles


    def _basestyle(self,icondict=None):
        """ Return default KML icon style """

        basestyle = skml.Style()
        basestyle.labelstyle.color = skml.Color.hex(self.colors['white'])
        basestyle.labelstyle.scale = 0.7
        basestyle.iconstyle.icon.href = self.iconshapes['square']
        basestyle.iconstyle.scale = 0.9
        basestyle.iconstyle.color= skml.Color.hex('33FF33')

        # only show desciption in balloon
        basestyle.balloonstyle.text = r'$[description]' 

        if icondict is not None:
            basestyle = self._changestyle(kmlstyle=basestyle,
                                          icondef=icondict)

        return basestyle


    def _changestyle(self,kmlstyle=None,icondef={}):
        """ Change existing kml style with definitions given in dictionary

        Parameters
        ----------
        kmlstyle : skml.style
            existing skml style object
        icondef : dict
            dictionary with items and values to change

        Returns
        -------
        KML style object

        """

        if not isinstance(kmlstyle,skml.styleselector.Style):
            msg = f'Variable {kmlstyle} must be a simplekml style object.'
            raise TypeError(msg)

        if not isinstance(icondef,dict):
            msg = f'Variable {icondef} must be a dictionary not {type(icondef)}.'
            raise TypeError(msg)

        for itemkey in icondef.keys():

            if itemkey=='iconshape':
                icon_name = icondef['iconshape']
                kmlstyle.iconstyle.icon.href = self.iconshapes[icon_name]

            elif itemkey=='iconscale':
                kmlstyle.iconstyle.scale = icondef['iconscale']

            elif itemkey=='iconcolor':
                mycolor = icondef['iconcolor']
                if mycolor.startswith('#'): 
                    mycolor = mycolor[1:]               
                kmlstyle.iconstyle.color= skml.Color.hex(mycolor)

            elif itemkey=='labelcolor':
                labelcolor = icondef['labelcolor']
                if labelcolor.startswith('#'): 
                    mycolor = mycolor[1:]               
                kmlstyle.labelstyle.color = skml.Color.hex(labelcolor)

            elif itemkey=='labelscale':
                kmlstyle.labelstyle.scale = float(icondef['labelscale'])

        return kmlstyle


    def _addcustomstyle(self,name,colorname=None,hexcolor=None,
        iconshape=None,styledic=None):
        """Add new custom KML waypoint style to customstyles"""

        newstyle = self.basestyle()

        if isinstance(styledic,dict):

            newstyle.iconstyle.icon.href = self.icons[styledic['iconshape']]
            newstyle.iconstyle.scale = styledic['iconscale']

            mycolor = styledic['iconcolor']
            if mycolor.startswith('#'): 
                mycolor = mycolor[1:]               
            newstyle.iconstyle.color= skml.Color.hex(mycolor)

            labelcolor = styledic['labelcolor']
            if labelcolor.startswith('#'): 
                mycolor = mycolor[1:]               
            newstyle.labelstyle.color = skml.Color.hex(labelcolor)
            newstyle.labelstyle.scale = float(styledic['labelscale'])

        if colorname is not None:

            if colorname in self.colors.keys(): 
                hexcolor = self.colors[colorname]

            if hexcolor.startswith('#'): 
                hexcolor = int(hexcolor[1:])

            newstyle.iconstyle.color= skml.Color.hex(hexcolor)

        if iconshape is not None:

            if iconshape in self.icons.keys(): 
                iconshape = self.icons[iconshape]
            else: 
                iconshape = self.icons['square']
            newstyle.iconstyle.icon.href = iconshape

        self.customstyles[name] = newstyle
        return self.customstyles[name]


    def writekml(self,kmlfile,metadata=True,cols=None):
        """ Write list of waypoints in RD to KML

        Parameters
        ----------
        kmlname : str
            name of kml file
        metadata : True,False, optional
            add metadata in bollontext (True) or not (False)
        cols : list of str, optional
            list of column names to add to balloontext of waypoints
        """
        #TODO: check if kmlfile path exists

        # Make list of fields to write to balloontext
        ##self.ballooncols = cols


        for idx, wp in self.wplist.iterrows():
   
            pnt = self.kml.newpoint(name=wp[self.label], 
                        coords=[(wp['lon'],wp['lat'])])  # lon, lat optional height
            pnt.snippet.maxlines = 0
            pointname = r'<font color='#000' size='+1'><strong>'+wp[self.label]+'</strong></font><br>'

            # assign waypoint style
            if (isinstance(self.stylecol,str) and isinstance(self.styledict,dict)):
                pnt.style = self.pointstyles[wp[self.stylecol]]
            else:
                pnt.style = self._basestyle(icondict=self.styledict)

            if metadata is True:

                if cols is None:
                    cols = list(wp.index)
                else:
                    cols = [x.lower() for x in cols]
                fields = zip(wp[cols].index, wp[cols].values)

                for ifield,field in enumerate(fields):
                    if ifield==0: 
                        tekst = field[0]+': '+str(field[1])+'<br>'
                    else: 
                        tekst = tekst + field[0]+': '+str(field[1])+'<br>'
            
                # save balloontext to kml
                pnt.description = tekst

        self.kml.save(kmlfile)

        return self.kml





import pathlib
import warnings
import numpy as np
from pandas import Series,DataFrame
import pandas as pd
from ..gwseries import GwSeries


class WaterWeb:
    """
    Manage WaterWeb dataset

    Methods
    -------
    from_csv
        Read waterweb csv export file and return WaterWeb object.
    srnames()
        Return list of series names.
    locname(srname)
        Return location name for given series.
    filname(srname)
        Return filter name for given series.
    seriestype
        Return type of measurement series for series srname.
    type_counts
        Return table of measurement type counts.
    networkname(name=None):
        Return or set network name.
    locprops(srname)
        Return series location properties.
    tubeprops(srname)
        Return welltube properties.
    levels(srname,ref='mp')
        Return measured levels for a series.
    gwseries(srname)
        Return gwseries object for a series.
    """
    _column_mapping = {
        'Lokatie':'sunloc',
        'SUN-kode':'sunsr',
        'NITG-kode':'nitgsr',
        'OLGA-kode':'olga',
        'DERDEN-kode':'derden',
        'X coordinaat':'xcr',
        'Y coordinaat':'ycr',
        'NAP hoogte bovenkant peilbuis':'mpnap',
        'Hoogte maaiveld tov NAP':'mvnap',
        'Hoogte maaiveld tov Nulpunt':'mvmp',
        'Hoogte maaiveld tov maaiveld':'mvmv',
        'NAP hoogte bovenkant filter':'filtopnap',
        'NAP hoogte onderkant filter':'filbotnap',
        'Peilmoment':'datetime',
        'Peilstand':'peilcmmp',
        'Peilstand in Meters':'peilmmp',
        'Peilstand tov NAP':'peilcmnap',
        'Peilstand tov NAP in Meters':'peilmnap',
        'Peilstand tov maaiveld':'peilcmmv',
        'Peilstand tov maaiveld in Meters':'peilmmv',
        'Peilkode':'peilcode',
        'Opmerking bij peiling':'peilopm'
        }

    _typedict = {
        'Lokatie':str,
        'SUN-kode':str,
        'NITG-kode':str,
        'OLGA-kode':str,
        'DERDEN-kode':str,
        'X coordinaat':'float64',
        'Y coordinaat':'float64',
        'NAP hoogte bovenkant peilbuis':'float64',
        'Hoogte maaiveld tov NAP':'float64',
        'Hoogte maaiveld tov Nulpunt':'float64',
        'Hoogte maaiveld tov maaiveld':'float64',
        'NAP hoogte bovenkant filter':'float64',
        'NAP hoogte onderkant filter':'float64',
        'Peilstand':str,
        'Peilstand':'float64',
        'Peilstand in Meters':'float64',
        'Peilstand tov NAP':'float64',
        'Peilstand tov NAP in Meters':'float64',
        'Peilstand tov maaiveld':'float64',
        'Peilstand tov maaiveld in Meters':'float64',
        'Peilkode':str,
        'Opmerking bij peiling':str,
        }

    _locprops_cols = ['sunloc','sunsr','nitgsr','olga','derden',
        'xcr','ycr',]

    _tubeprops_cols = ['mpnap','mvnap','filtopnap','filbotnap']

    _peilprops_cols = ['datetime','peilcmmp','peilcode','peilopm']

    _locprops_mapping = {
        'locname':'sunloc','alias':'nitgsr',
        'xcr':'xcr','ycr':'ycr'
        }

    _tubeprops_mapping = {
        'startdate':'datetime','mplevel':'mpnap',
        'filtop':'filtopnap','filbot':'filbotnap',
        'surfacelevel':'mvnap'
        }

    _levels_mapping = {
        'headdatetime':'datetime', 'headmp':'peilcmmp', 
        'headnote':'peilcode','remarks':'peilopm'}

    _reflevels = [
        'datum','surface','mp',
        ]

    _measurement_types = ['B','S','L','P']


    def __init__(self,fpath=None,data=None,network=None):

        self._fpath = fpath
        self._network = network
        self._data = data

        if ((self._fpath is not None) and (self._data is None)):

            if not pathlib.Path(self._fpath).is_file():
                raise ValueError((f'{self._fpath} is not a valid '
                    'file path.'))

            self._data = self._readcsv(self._fpath)

        if self._network is None:
            if self._fpath is not None:
                path = pathlib.Path(fpath)
                self._network = path.stem
            else:
                self._network = 'waterweb'

        if self._data is not None:

            if not isinstance(self._data,pd.DataFrame):
                raise ValueError((f'{self._data} is not a valid Pandas ')
                    (f'DataFrame.'))

    def __repr__(self):
        return (f'{self._network} (n={self.__len__()})') #len(self.srnames())})')


    def __len__(self):
        return len(self.srnames())


    def _readcsv(self,fpath):
        """Read WaterWeb csv export file

        Parameters
        ----------
        fpath : str
            valid file path to WaterWeb csv export file

        Returns
        -------
        pd.DataFrame   
        
        """

        file = open(fpath,'r')
        self._flines = file.readlines()
        file.close()

        # read flines to list of lists
        datalines = []
        for rownr,line in enumerate(self._flines):

            line = line.rstrip('\n').rstrip(';')

            if rownr==0:
                colnames = [x.strip('"') for x in line.split(';')]
                continue

            elms = [x.strip('"') for x in line.split(';')]
            elms = [x.replace(',','.') for x in elms]

            elms = [x if len(x)>1 else np.nan for x in elms]

            if len(elms)>1:
                datalines.append(elms)

        # create dataframe and do type conversions
        data = pd.DataFrame.from_records(datalines,columns=colnames)
        data = data.astype(dtype=self._typedict)
        data = data.rename(columns=self._column_mapping)

        # modify columns
        data['datetime'] = pd.to_datetime(data['datetime'])
        data['nitgsr'] = data['nitgsr'].apply(
            lambda x:x[:8]+"_"+x[-3:].lstrip('0'))

        return data


    @classmethod
    def from_csv(cls,fpath,network=None):
        """ 
        Read waterweb csv network file and return new WaterWeb object

        Parameters
        ----------
        filepath : str
            path to waterweb csv export file

        networkname : str, optional
            name of network

        Returns
        -------
        WaterWebNetwork object

        """
        data = cls._readcsv(cls,fpath)
        return cls(fpath=fpath,data=data,network=network)


    def srnames(self):
        """Return list of series names"""
        return self._data['sunsr'].unique()

    def seriestype(self,srname=None):
        """Return type of measurement series

        Parameters
        ----------
        srname : str, optional
            series name, if not given, type of all series is returned

        Returns
        -------
        str, numpy array of str """

        srnames = self.srnames()
        sr = Series(data=srnames,index=srnames,name='seriestype')
        sr = sr.apply(lambda x:x[8])

        if srname is not None:
            return sr[srname]

        return sr


    def type_counts(self):
        """Return table of measurement type counts."""
        srtypelist = []
        for name in self.srnames():
            srtypelist.append(self.seriestype(name))
        tbl = pd.Series(srtypelist).value_counts()
        tbl.name = self.networkname()
        return tbl


    def locname(self,srname):
        """Return location name for given series

        Parameters
        ----------
        srname : str
            name of series to return """

        return self.locprops(srname)['sunloc']


    def filname(self,srname):
        """Return filter name for given series

        Parameters
        ----------
        srname : str
            name of series to return 

        Notes
        -----
        WaterWeb uses the DINO-SUN convention of only explicitly naming
        filters if more than one filter is present."""

        filname = self.locprops(srname)['sunsr']
        if filname[-1].isalpha():
            return filname[-1]
        return ''


    def networkname(self,name=None):
        """Return or set network name

        Parameters
        ----------
        name : str, optional
            name of measurement network """

        if isinstance(name,str):
            self._network = name
        return self._network


    def locprops(self,srname):
        """Return series location properties

        Parameters
        ----------
        sunsr : str
            name of series to return

        Return
        ------
        pd.Series """

        data = self._data[self._data['sunsr']==srname]
        lastrow = data.iloc[-1,:]

        sr = Series(index=self._locprops_cols,dtype='object',
            name=srname)
        for col in self._locprops_cols:
            sr[col] = lastrow[col]

        return sr


    def tubeprops(self,srname):
        """Return welltube properties

        Parameters
        ----------
        sunsr : str
            name of series

        Returns
        -------
        pd.DataFrame """

        data = self._data[self._data['sunsr']==srname]
        data = data.drop_duplicates(subset=self._tubeprops_cols,
            keep='first')
        data = data[['datetime']+self._tubeprops_cols]
        data = data.reset_index(drop=True)

        return data

    def levels(self,srname,ref='mp'):
        """Return measured levels

        Parameters
        ----------
        srname : str
            name of series
        ref : {'mp','datum',' surface'}, default 'mp' 

        Returns
        -------
        pd.Series """

        if ref not in self._reflevels:
            warnings.warn((f'{ref} not in {self._references}. '),
                (f'reference is set to "datum".')) 
            ref = 'datum'
        if ref=='mp':
            col = 'peilcmmp'
        if ref=='datum':
            col = 'peilmnap'
        if ref=='mv':
            col = 'peilcmmv'

        data = self._data[self._data['sunsr']==srname]
        data = data[[col,'datetime']]
        data = data.set_index('datetime',drop=True).squeeze()
        data.name = self.locname(srname)
        data.index.name = 'datetime' 
        data = data/100.

        return data


    def gwseries(self,srname):
        """Return gwseries obect for one series

        Parameters
        ----------
        srname : str
            name of series to return

        Returns
        -------
        acequia.GwSeries

        """

        gw = GwSeries()

        # locprops
        locprops = self.locprops(srname)
        for gwprop in list(gw._locprops.index):
            if gwprop not in self._locprops_mapping.keys():
                continue
            wwnprop = self._locprops_mapping[gwprop]
            gw._locprops[gwprop] = locprops[wwnprop]

        gw._locprops['filname'] = self.filname(srname)
        gw._locprops['height_datum'] = 'mNAP'
        gw._locprops['grid_reference'] = 'RD'

        # tubeprops
        tubeprops = self.tubeprops(srname)
        for gwprop in list(gw._tubeprops):
            if gwprop not in self._tubeprops_mapping.keys():
                continue
            wwnprop = self._tubeprops_mapping[gwprop]
            gw._tubeprops[gwprop] = tubeprops[wwnprop].values
            if gwprop in gw._tubeprops_numcols:
                gw._tubeprops[gwprop] = gw._tubeprops[gwprop]/100.

        #levels
        levels = self._data[self._data['sunsr']==srname]
        levels = levels[self._peilprops_cols]
        for gwprop in list(gw._headprops_names):
            if gwprop not in self._levels_mapping.keys():
                continue
            wwnprop = self._levels_mapping[gwprop]
            gw._heads[gwprop] = levels[wwnprop].values
        gw._heads['headmp'] = gw._heads['headmp']/100.

        return gw




""" This module contains a class Quantiles that calculates quantiles
of measured heads for hydrological years.

Author: Thomas de Meij

"""

from datetime import datetime
import datetime as dt
import warnings
import numpy as np
from pandas import Series, DataFrame
import pandas as pd
import matplotlib.pyplot as plt

#import acequia as aq
from acequia.gwseries import GwSeries


class Quantiles:
    """Calculate quantiles from series of measured heads

    Parameters
    ----------
    ts : pd.Series, aq.GwSeries
        timeseries with groundwater head measurments

    ref : str, ['datum','surface']
        reference level for measurements

    nclasses : int (default 10)
        number of quantile classes 

    days : bool, default True
        label xax with days or quantiles

    """

    n14 = 18

    def __init__(self, gw, srname=None, ref='surface', nclasses=10, days=True):
        """Return quantiles object"""

        if isinstance(gw,aq.GwSeries):
            ts = gw.heads(ref=ref)
            if srname is None:
                srname = gw.name()

        if isinstance(gw,pd.Series):
            ts = gw
            if srname is None: 
                srname = gw.name

        if srname is None: 
            srname = 'series'

        self.gw = gw
        self.ref = ref
        self.ts = ts
        self.srname = srname
        self.days = days

        if self.days:
            
            self.days = [x*30 for x in range(12)] + [365]
            self.qt = [x/365 for x in self.days]
            self.qtlabels = [str(x) for x in self.days]
            """
            self.days = [x*5 for x in range(74)] #+ [365]
            self.qt = [x/365 for x in self.days]
            self.qtlabels = [str(x) for x in self.days] # if (x/30).is_integer()]
            """

        else:
            self.qt = np.linspace(0,1,nclasses+1) # list of quantiles
            self.qtlabels = ['p'+str(int(x*100)) for x in self.qt]
            self.days = [int(x*365) for x in self.qt]

        self.tbl = self._quantiles()


    def _quantiles(self):
        """Return table with quantiles for each hydrological year"""

        # empty table with hydroyears and percentiles
        hydroyear = aq.hydroyear(self.ts)
        #allyears = np.arange(hydroyear.min(),hydroyear.max()+1)
        tbl = pd.DataFrame(index=set(hydroyear),columns=self.qtlabels)


        for i,(name,val) in enumerate(zip(self.qtlabels,self.qt)):
            grp = self.ts.groupby(hydroyear)
            tbl[name] = grp.quantile(val)

        return tbl


    def table(self):
        """Return table of quantiles for each hydrological year"""
        return self.tbl


    def plot(self,years=None,figpath=None,figtitle=None,ylim=None,
        ignore=None):
        """Plot quantiles

        Parameters
        ----------
        years : list of int, optional
            years to plot in color
        figpath : str, optional
            figure output path
        figtitle : str, optional
            figure title
        ylim : list, optional
            ymin, ymax
        ignore : list, optional
            years to ignore in calculating reference
        """

        if years is None:
            years = []

        if not isinstance(years,list):
            msg = f'years of type {type(list)} changed to empty list'
            warnings.warn(msg)
            years = []

        if figtitle is None:
            figtitle = self.srname

        csurf = '#8ec7ff'
        clines = '#2f90f1' #'#c1e0ff'
        cyears = '#b21564'

        fig,ax = plt.subplots(1,1)
        fig.set_figwidth(13)
        fig.set_figheight(7)

        tbl = self.tbl
        x = self.qt

        reftbl = tbl.copy()
        if ignore:
            idx = [x for x in reftbl.index.values if x not in ignore]
            reftbl = reftbl.loc[idx,:]

        # reference based on quantiles
        upper = reftbl.quantile(0.05)*100
        lower = reftbl.quantile(0.95)*100

        # alternative reference
        #mean = reftbl.median(axis=0,skipna=True)
        #std = reftbl.std(axis=0,skipna=True)
        #upper = (mean+std)*100
        #lower = (mean-std)*100

        # alternative reference 2
        # leave out lowest line
        #upper = [reftbl[col].sort_values()[:-1].quantile(0.05)*100 for col in reftbl.columns]
        #lower = [reftbl[col].sort_values()[:-1].quantile(0.95)*100 for col in reftbl.columns]

        ax.fill_between(x, upper, lower, color=csurf) 

        for year in self.tbl.index:

            yvals = self.tbl.loc[year,:].values * 100
            xvals = self.qt

            #if year in years:
            #    ax.plot(xvals,yvals,color=cyears)
            #else:
            ax.plot(xvals,yvals,color=clines)


        for year in self.tbl.index:
            if year in years:
                yvals = self.tbl.loc[year,:].values * 100
                xvals = self.qt
                ax.plot(xvals,yvals,color=cyears)

        ax.set_xticks(self.qt)
        ax.set_xticklabels(self.qtlabels)
        if self.days:
            ##ax.set_xticklabels(self.days)
            ax.set_xlabel('dagen/jaar', fontsize=15)
        else:
            ax.set_xlabel('percentiel', fontsize=15)

        ax.set_xlim(1,0)
        if ylim:
            ax.set_ylim(ylim[0],ylim[1])

        ax.invert_xaxis()
        if self.ref=='surface':
            ax.invert_yaxis()

        ax.set_ylabel('grondwaterstand (cm -mv)', fontsize=15)


        ax.text(0.99, 0.97, figtitle, horizontalalignment='right',
                verticalalignment='top', transform=ax.transAxes,
                fontsize = 16)

        if figpath is not None:
            plt.savefig(figpath,bbox_inches='tight')

        plt.show()
        return ax



"""This module contains the HydroMonitor object for reading groundwater
head measurements from a HydroMonitor csv export file

"""

from collections import OrderedDict
import warnings
import os.path
import errno
import os

import matplotlib as mpl
import matplotlib.pyplot as plt
from pandas import Series, DataFrame
import pandas as pd
import numpy as np

import acequia as aq


class HydroMonitor:
    """Read and manage data from hydromonitor csv export file
    
    Parameters:
    ---------
    header : pandas dataframe, optional
        header data from hydromonitor file as pandas dataframe
    metadata : pandas dataframe, optional
        metadata data from hydromonitor file as pandas dataframe
    data : pandas dataframe, optional
        measurements from hydromonitor file as pandas dataframe

    Examples:
    -------
    Read hydromonitor csv export file:
    >>>hm = HydroMonitor.from_csv(filepath=<path>)
    Convert to list of GwSeries objects:
    >>>mylist = hm.to_list()
    Iterate over all series and return GwSeries objects one at a time:
    >>>for i in range(len(hm)):
        gw = next(hm)
    Iterate over raw data and returnGwSeries objects: 
    >>>for (loc,fil),data in hm.iterdata():
        gw = hm.get_series(data=data,loc=loc,fil=fil)
    Save all series to json files in <filedir>:
    >>>hm.to_json(<filedir>)

    """

    CSVSEP = ";"
    METATAG = 'StartDateTime' #;XCoordinate;YCoordinate;'
    DATATAG = 'DateTime' #;LoggerHead;ManualHead;'


    mapping_tubeprops = OrderedDict([
        ('startdate','startdatetime'),
        ('mplevel','welltoplevel'),
        ('filtop','filtertoplevel'),
        ('filbot','filterbottomlevel'),
        ('surfacedate',''),
        ('surfacelevel','surfacelevel'),
        ])


    def __repr__(self):
        return (f'{self.__class__.__name__} instance')
      

    def __init__(self,header=None,metadata=None,data=None):
        """ Initialize HydroMonitor object """

        if isinstance(header, pd.Series):
            self.header = header
        else:
            if not header:
                self.header = DataFrame()
            else:
                message = 'header must be type None or type DataFrame'
                raise TypeError(message)

        if isinstance(metadata, pd.DataFrame):
            self.metadata = metadata
        else:
            if not metadata:
                self.metadata = DataFrame()
            else:
                message = 'meta must be type None or type DataFrame'
                raise TypeError(message)

        if isinstance(data, pd.DataFrame):
            self.data = data
        else:
            if not data:
                self.data = DataFrame()
            else:
                message = 'data must be type None or type DataFrame'
                raise TypeError(message)

        # create generator
        data = self.delete_duplicate_data()
        self.srgen =  data.groupby(self.idkeys()).__iter__()
        self.itercount = 0

    @classmethod
    def from_csv(cls,filepath):
        """ 
        read menyanthes hydromonitor csv export file

        parameters
        ----------
        filepath : str
            path to hydromonitor csv export file

        returns
        -------
        HydroMonitor object

        """
        header,metadata,data = cls._readcsv(cls,filepath)
        return cls(header=header,metadata=metadata,data=data)


    def _readcsv(self,filepath):
        """ Read hydromonitor csv export file """
        
        try:
            self.filepath = filepath
            textfile = open(self.filepath)
        except IOError:
            raise FileNotFoundError(
                errno.ENOENT, os.strerror(errno.ENOENT), filepath)
            textfile = None
        else:
            self.header,self.line_numbers = self._read_header(self,textfile)
            textfile.close()

            content_list = [x.lower() for x in self.header['file_contents']]
            if 'metadata' in content_list:
                self.metadata = self._read_metadata(self)
            else:
                self.metadata = None
                warnings.warn("Metadata not specified.")

            if 'data' in content_list:
                if self.metadata is None:
                    raise(f'Heads from {filepath} can not be read ',
                          f'because metadata are not available.')                
                self.data = self._read_data(self)
            else:
                self.data = None
                warnings.warn("Data not specified.")
        #finally:

        return self.header,self.metadata,self.data


    def _read_header(self,textfile):
        """ 
        read header and linenumbers from hydromonitor export file 

        returns
        -------
        header : pandas dataframe
            header items as pandas dataframe
        filelines : tuple
            line numbers

        """

        metatag_found = False
        datatag_found = False

        header = OrderedDict()
        for i,line in enumerate(textfile):

            line = line.rstrip()
            linelist = line.split(self.CSVSEP)

            # read header tags
            if line.startswith('Format Name'):
                header['format_name']=linelist[1]
            elif line.startswith('Format Version'):
                header['format_version']=linelist[1]
            elif line.startswith('Format Definition'):
                header['format_definition']=linelist[1]
            elif line.startswith('File Type'):
                header['file_type']=linelist[1]
            elif line.startswith('File Contents'):
                values = list(filter(None, linelist[1:]))
                header['file_contents']=values
            elif line.startswith('Object Type'):
                header['object_type']=linelist[1]
            elif line.startswith('Object Identification'):
                values = list(filter(None, linelist[1:]))
                header['object_identification']=values

            # read metadata line number and column names
            elif self.METATAG in line:
                metatag_found = True
                self.metafirst = i+2
                self.metacols = [x.lower() for x in linelist]

            # read data line number column names
            elif self.DATATAG in line:
                datatag_found = True
                self.datafirst = i+2
                self.metalast = i-2
                self.datacols = [x.lower() for x in linelist]
                break # avoid iterating over lines after metadata

        # warnings
        if not metatag_found:
            msg = f'Metadata header {self.METATAG} not found.'
            warnings.warn(msg)

        if not datatag_found:
            msg = f'Data header {self.DATATAG} not found.'
            warnings.warn(msg)

        # return variables
        self.header = Series(header)
        self.line_numbers = (self.metafirst,self.metalast,self.datafirst)
        return self.header,self.line_numbers


    def _read_metadata(self):
        """ read metadata from hydromonitor csv export file """

        #Name;NITGCode;OLGACode;FilterNo;StartDateTime;XCoordinate;
        #YCoordinate;SurfaceLevel;WellTopLevel;FilterTopLevel;
        #FilterBottomLevel;WellBottomLevel;LoggerSerial;LoggerDepth;
        #Comment;CommentBy;Organization;Status;

        meta_nrows = self.metalast - self.metafirst + 1
        dfmeta = pd.read_csv(
            self.filepath,
            sep=self.CSVSEP,
            index_col=False,
            header=None,
            names=self.metacols,
            skiprows=self.metafirst,
            nrows=meta_nrows,
            dtype=str,
            encoding='latin-1',
            )

        # delete empty last column
        if '' in list(dfmeta.columns):
            dfmeta = dfmeta.drop([''], axis=1)

        dfmeta['startdatetime'] = pd.to_datetime(dfmeta['startdatetime'],
                                  format='%d-%m-%Y %H:%M',
                                  errors='coerce')

        numcols = ['xcoordinate','ycoordinate','surfacelevel',
                   'welltoplevel','filtertoplevel','filterbottomlevel',

                   'wellbottomlevel',]
        for col in numcols:
            dfmeta[col] = pd.to_numeric(dfmeta[col],errors='coerce')

        return dfmeta

    def _read_data(self):
        """ read data from hydromonitor csv export file """

        #read data
        dfdata = pd.read_csv(
            self.filepath,
            sep=self.CSVSEP,
            index_col=False,
            header=None,
            names=self.datacols,
            skiprows=self.datafirst,
            #parse_dates=['datetime'], # don't, this takes a lot of time
            dtype=str,
            encoding='latin-1',
            )

        if 'loggerhead' not in self.datacols:
        # when no loggerhead is available, menyanthes only exports
        # the column manualheads and the column loggerheads is simply missing
        # this happens when loggerdata without manual control measurments
        # are imported from another source; Menyanthes marks these 
        # measurements as manual heads.

            pos = dfdata.columns.get_loc('datetime')+1
            dfdata.insert(loc=pos,column='loggerhead',value=np.nan)

            msg = f'Missing data column loggerhead added and filled with NaNs'
            warnings.warn(msg)

        if 'manualhead' not in self.datacols:
        # this is a variation on the previous missing loggerhead issue

            pos = len(dfdata.columns)
            dfdata.insert(loc=pos,column='manualhead',value=np.nan)
            #dfdata['manualhead'] = np.nan

            msg = f'Missing data column loggerhead added and filled with NaNs'
            warnings.warn(msg)


        # delete empty last column
        if '' in list(dfdata.columns):
            dfdata = dfdata.drop([''], axis=1)

        # delete repeating headers deep down list of data as a result of
        # annoying bugs in menyanthes export module
        dfdata = dfdata[dfdata.nitgcode!='NITGCode'].copy()
        dfdata = dfdata[dfdata.nitgcode!='[String]'].copy()

        # parsing these dates is very time consuming
        dfdata['datetime'] = pd.to_datetime(dfdata['datetime'],
              dayfirst=True,format='%d-%m-%Y %H:%M',errors='coerce')
        dfdata['loggerhead'] = pd.to_numeric(dfdata['loggerhead'], 
                               errors='coerce')
        dfdata['manualhead'] = pd.to_numeric(dfdata['manualhead'], 
                               errors='coerce')
        return dfdata


    def idkeys(self):
        """Return column names that give a unique identification of a 
        series """
        # bug in hydromonitor export? id is allways nitgcode, filterno
        # idkeys = ['nitgcode','filterno']
        if len(self.header['object_identification'])>2:
            message.warn('More than two object identification keys given')
        return [x.lower() for x in self.header['object_identification']]


    def delete_duplicate_data(self):
        """Remove duplicate data from data and return pd.DataFrame

        Duplicates occur in groundwater head measurments in
        hydromonitor export when loggervalue and manual control 
        measurement have the same timestamp."""
        sortcols = self.idkeys() + ['datetime','manualhead']
        data = self.data.sort_values(by=sortcols)
        dropcols = self.idkeys() + ['datetime']
        self.data_no_dups = data.drop_duplicates(subset=dropcols, 
                            keep='first').copy()
        return self.data_no_dups


    def get_series(self,data=None,loc=None,fil=None):
        """Return GwSeries object from HydroMonitor object

        Parameters
        ----------
        data : pd.Dataframe
            Table with groundwater head values
        loc : str
            Well location name
        fil : str
            Tube name

        Returns
        -------
        GwSeries object

        """

        gws = aq.GwSeries()

        # select metadata for series 
        bool1 = self.metadata[self.idkeys()[0]]==loc
        bool2 = self.metadata[self.idkeys()[1]]==fil
        metadata = self.metadata[bool1&bool2]
        firstindex = metadata.index[0]

        # set tubeprops from hydro,onitor metadata
        for prop in list(gws._tubeprops):
            metakey = self.mapping_tubeprops[prop]
            if len(metakey)>0:
                gws._tubeprops[prop] = metadata[metakey].values
            if prop not in self.mapping_tubeprops.keys():
                warnings.warn(f"Unknown property {prop} in {type(gws)}")

        # set locprops from hydro,onitor metadata
        for prop in list(gws._locprops.index):

            if prop=='locname':
                gws._locprops[prop] = metadata.at[firstindex,self.idkeys()[0]]
            
            if prop=='filname':
                gws._locprops[prop] = metadata.at[firstindex,self.idkeys()[1]]

            if prop=='alias':
                if 'nitgcode' in self.idkeys(): 
                    alias_key = 'name'
                else: 
                    alias_key = 'nitgcode'
                gws._locprops[prop] = metadata.at[firstindex,alias_key]

            if prop=='xcr':
                gws._locprops[prop] = metadata.at[firstindex,'xcoordinate']

            if prop=='ycr':
                gws._locprops[prop] = metadata.at[firstindex,'ycoordinate']

            if prop=='height_datum':
                gws._locprops[prop] = 'mnap'

            if prop=='grid_reference':
                gws._locprops[prop] = 'rd'

            if prop not in ['locname','filname','alias','xcr','ycr',
                            'height_datum','grid_reference']:

                warnings.warn(f"Unknown property {prop} in {type(gws)}")


        # set gwseries
        datetimes = data['datetime'].values
        heads = np.where(
                    np.isnan(data['loggerhead'].values),
                    data['manualhead'].values,
                    data['loggerhead'].values)
        heads = Series(data=heads,index=datetimes)

        # remove rows with invalid datevalues and warn
        NaTdate = pd.isna(heads.index)
        nNaT = len(heads[NaTdate])
        if nNaT!=0:
            msg = f"Removed {nNaT} rows with invalid date from {gws.name()}"
            warnings.warn(msg)
            heads = heads[~NaTdate].copy()

        # convert heads in mnap to mref
        gws._heads = heads
        for index,props in gws._tubeprops.iterrows():
            mask = gws._heads.index >= props['startdate'] 
            gws._heads = gws._heads.mask(
                         mask,props['mplevel']-heads)

        return gws

    def to_list(self):
        """ Return data from HydroMonitor as a list of GwSeries() 
            objects """

        srlist = []

        heads = self.delete_duplicate_data()
        filgrp = heads.groupby(self.idkeys())
        for (location,filnr),data in filgrp:

            gws = self.get_series(data=data,loc=location,fil=filnr)
            srlist.append(gws)

        return srlist


    def __iter__(self):
        return self


    def __next__(self):

        if self.itercount >= len(self):
            raise StopIteration

        (loc,fil),data = next(self.srgen)
        gw = self.get_series(data=data,loc=loc,fil=fil)

        self.itercount+=1
        return gw

    def iterdata(self):
        """Return generator for iterating over heads data"""
        heads = self.delete_duplicate_data()
        return heads.groupby(self.idkeys()).__iter__()

    def __len__(self):
        heads = self.delete_duplicate_data()
        hymlen=0
        for srname,sr in heads.groupby(self.idkeys()).__iter__():
            hymlen+=1
        return hymlen

    def to_json(self,filedir=None):

        for (loc,fil),data in self.iterdata():
            gws = self.get_series(data=data,loc=loc,fil=fil)
            gws.to_json(filedir)




from pathlib import Path
import warnings
import numpy as np
import pandas as pd
from pandas import DataFrame, Series


class KnmiWeather:
    """Read weather data from knmi file

    Parameters
    ----------
    fpath : str
        path to knmi txt file with weather data


    Examples
    --------

    wth = aq.KnmiWeather(<filepath>)

    evp = wth.timeseries('evp')

    name = wth.name()


    """

    NAMESTR = ''.join([
        'STN,YYYYMMDD,DDVEC,FHVEC,   FG,  FHX, FHXH,  FHN, FHNH,',
        '  FXX, FXXH,   TG,   TN,  TNH,   TX,  TXH, T10N,T10NH,   SQ,',
        '   SP,    Q,   DR,   RH,  RHX, RHXH,   PG,   PX,  PXH,   PN,',
        '  PNH,  VVN, VVNH,  VVX, VVXH,   NG,   UG,   UX,  UXH,   UN,',
        '  UNH, EV24'])

    VARNAMES = ['prc','evp','rch']

    SKIPROWS = 48

    def __init__(self,fpath):


        self.colnames = self.NAMESTR.replace(' ','').split(',')
        self.keepcols = 'YYYYMMDD,RH,EV24'.split(',')
        self.fpath = Path(fpath)


        if not (self.fpath.exists() and self.fpath.is_file()):
            msg = [f'Filepath \'{self.fpath}\' not found',
                'empty dataframe is returned.']
            warnings.warn(' '.join(msg))
            self.rawdata = pd.DataFrame(columns=self.colnames)
            self.data = pd.DataFrame(columns=self.keepcols)
        else:
            self._readfile()
            self._clean_rawdata()


        self.stn = self.rawdata.loc[0,'STN']


    def _readfile(self):

        # read csv to pd.DataFrame with only str values
        self.rawdata = pd.read_csv(self.fpath,sep=',',
            skiprows=self.SKIPROWS,
            names=self.colnames,dtype='str')

        # replace empty strings with NaN
        self.rawdata = self.rawdata.apply(
            lambda x: x.str.strip()).replace('', np.nan)


    def _clean_rawdata(self):

        self.data = self.rawdata[self.keepcols].copy()
        for colname in list(self.data):

            if colname=='YYYYMMDD':

                # create datetimeindex from string column
                self.data[colname] = pd.to_datetime(self.data[colname],
                    infer_datetime_format=True)
                self.data = self.data.set_index(
                    colname,verify_integrity=True)

                # make sure all dates are in index
                firstdate = self.data.index[0]
                lastdate = self.data.index[-1]
                idx = pd.date_range(start=firstdate, end=lastdate,
                    freq='D')
                self.data = self.data.reindex(idx)
                self.data.index.name='date'

            if colname=='RH':
                # RH is precipitation in 0.1 mm/day
                # RH = -1 means RH < 0.05 mm/day
                self.data[colname] = self.data[colname].replace('-1','0.5')

            if colname in ['RH','EV24']:
                self.data[colname] = self.data[colname].astype(float)/10.


    def timeseries(self,varname):
        """Return timeseries with data

        Parameters
        ----------
        varname : {'prc','evp','rch')
            variable to return
        """
 
        if varname not in self.VARNAMES:
            msg = [f'{varname} is not a valid variable name.',
                f'parameter varname must be in {self.VARNAMES}',
                f'by default rain data are returned']
            warnings.warn(' '.join(msg))
            varname = 'prc'

        if varname == 'prc':
            sr = self.data['RH']
            sr.name = 'prc'
        if varname == 'evp':
            sr = self.data['EV24']
            sr.name = 'evp'
        if varname == 'rch':
            sr = self.data['RH']-self.data['EV24']
            sr.name = 'rch'

        first = sr.first_valid_index()
        last = sr.sort_index(ascending=False).first_valid_index()
        return sr[first:last]


    def units(self):
        """Return table with definitions and units of variables"""

        tbl = pd.DataFrame({
            'variable' : ['prc','evp','rch'],
            'datacol' : ['RH','EV24','calculated'],
            'unit' : ['mm/day','mm/day','mm/day'], 
            })
        tbl = tbl.set_index('variable')
        return tbl





""" 
This module contains the class HeadsDiff for calculating and plotting 
differences between multiple groundwater head series.
"""

import warnings
import math

from pandas import Series, DataFrame
import pandas as pd

import matplotlib.pyplot as plt
import seaborn as sns; sns.set()
import numpy as np


def headsdif_from_gwseries(heads=None,locname=None,refcol=None):
    """Return HeadsDif object with data or None for invalid data.

    Parameters
    ----------
    heads : list of GwSeries objects
        Measured heads of multiple groundwater series.
    locname : str
        Location name for annotating graphs.
    refcol : str, optional
        Series name to use as reference.

    Returns
    -------
    HeadsDiff
        HeadsDif object or None

    Examples
    --------
    >>>hd = headsdif_from_gwseries(heads=<heads>,locname=<locname>,
            refcol=<refcol>)
    """
    if not isinstance(heads,list):
        msg = 'Parameter heads must be a list of GwSeries onjects'
        warnings.warn(msg)
        return None

    if not all([isinstance(x,acequia.gwseries.GwSeries) for x in heads]):
        msg = 'Parameter heads must be a list of GwSeries onjects'
        warnings.warn(msg)
        return None

    valid_heads = [x for x in heads if not x.heads().isnull().all()]
    if not len(valid_heads)==len(heads):
        removed = list(set(heads).difference(valid_heads))
        msg = f'Series {removed} with only NaNs removed from list'
        warnings.warn(msg)

    if len(valid_heads)<2:
        msg = 'Less than two heads series left after validation.'
        warnings.warn(msg)
        return None

    return HeadsDif(heads=valid_heads,locname=locname,refcol=refcol)


class HeadsDif:
    """Calculate head differences and descriptive statistics for 
    multiple groundwater head series

    Methods
    -------
    table_headsdif
        Return table with head differences relative to series ref.
    table_headsref
        Return table with heads relative to mean of entire series.
    difsums
        Return table with head differences by season.
    date_seasons
        Return list with seasons for each datetime in index.
    plot_time
        Plot al heads in one graph and all head differences below that 
        in one figure.
    plot_head
        Plot head difference by reference head value.
    plot_freq
        Plot head differences as grid of frequency plots.

    Notes
    -----
    Class HeadsDif requires input of a list of multiple GwSeries objects
    or pd.Series objects with valid head data. No checking is performed.

    To check input data before creating a HeadsDif object, us the custom
    function headsdif_from_gwseries(). This function returns None when 
    no valid heads could be calculated. 

    Examples
    --------
    Create headdiff object without checking input:
    >>>hd = acequia.headsdif(heads=<list>,locname=<str>, refcol=<str>)

    Call function headsdif_from_gwseires() and clean up data before
    creating a headsdiff object:
    >>>hd = headsdif_from_gwseries(heads=<heads>,locname=<locname>,
            refcol=<refcol>)
    >>>if hd is None:
    >>>     continue
    """
    _period_names = {'quarter','half-year'}

    def __init__(self,heads=None,locname=None,refcol=None):
        """
        Parameters
        ----------
        heads : list,pd.DataFrame
            Measured heads of multiple groundwater series.
        locname : str
            Location name for annotating graphs.
        refcol : str
            Series name to use as reference.
        """
        if heads is None:
            msg = f'Parameter heads must be given'
            raise ValueError(msg)

        if isinstance(heads,list):

            if isinstance(heads[0],acequia.GwSeries):
                heads = [gw for gw in heads if not 
                         gw.heads().isnull().all()]

                tslist = [gw.heads(ref='datum',freq='D')
                          for gw in heads]

                if locname is None:
                    self.locname = heads[0].locname()

            if isinstance(heads[0],pd.Series):
                tslist = heads

            self._heads = pd.concat(tslist, axis=1)

        if isinstance(heads,pd.DataFrame):
            self._heads = heads

        if locname is None:
            self.locname = self._heads.columns[0].split('_')[0]
        self.locname = locname

        if refcol is None:
            refcol = list(self._heads)[0]
        self._refcol = refcol

        # create standard variables
        self._check_heads()
        self._headsdif = self.table_headsdif()
        self._headsref = self.table_headsref()

    def __repr__(self):
        nsr = len(self.table_headsdif().columns)
        return (f'{self.__class__.__name__}(n={nsr})')

    def table_headsdif(self,ref=None):
        """
        Return table with head differences relative to series ref.

        If ref is not set, first series is taken as reference.

        Returns
        -------
        pd.DataFrame
        """
        if ref is None:
            ref = self._refcol

        if (ref not in self._heads.columns) and (ref is not None):
            msg = f'Reference column {ref} not in heads table.'
            warnings.warn(msg)

        heads = self._heads.copy()
        hdif = self._heads.copy()
        for i,col in enumerate(list(hdif)):
            hdif[col] = heads[col]-heads[ref]

        return hdif

    def table_headsref(self):
        """Return table with heads relative to mean of entire series."""
        heads = self._heads.copy()
        for i,col in enumerate(list(heads)):
            heads[col] = heads[col]-heads[col].mean()
        return heads

    def difsums(self,period='quarter'):
        """
        Return table with head differences by season.

        Parameters
        ----------
        seasons : {'quarter','half-year'}
            Aggregate by quarter or by year.

        Returns
        -------
        pd.DataFrame
        """
        if period not in self._period_names:
            msg = ''.join(
                  f'{period} is not a valid period name. ',
                  f'Period must be in {self._period_names}.',)
            raise ValueError(msg)

        seasons = self.date_seasons(self._headsdif,period=period)
        difsns = (self._headsdif*100).groupby(seasons).mean().round(2).T
        difsns.columns.name = None
        difsns.index.name = 'series'
        return difsns

    def _check_heads(self):
        """
        Check heads measurement table for errors, repair and show
        warnings.
        """ 

        for col in self._heads.columns:
            allnull = self._heads[col].isnull().all()
            if allnull:
                self._heads = self._heads.drop(columns=[col])

        refcol_missing = self._refcol not in self._heads.columns
        ncols = len(self._heads.columns)
        if refcol_missing and ncols>1:
            old_refcol = self._refcol
            self._refcol = self._heads.columns[0]

            msg = f'Dropped reference column {old_refcol}, replaced with {self._refcol}'
            warnings.warn(msg)

        if ncols < 2:
            msg = f'Less than two columns with valid data in {self.locname}'
            raise ValueError(msg)


    def date_seasons(self,dtindex,period='quarter'):
        """
        Return list with seasons for each datetime in index.

        Parameters
        ----------
        dtindex : pd.Datarame,pd.DateTimeIndex
            Index with dates.

        seasons : {'quarter','half-year'}
            Aggregate by season or by six months.

        Returns
        -------
        list
            Season for each datetime in dtindex
        """
        if isinstance(dtindex,pd.DataFrame):
            dtindex = dtindex.index

        if period=='quarter':
            seasons = ['1_Winter', '1_Winter', '2_Spring', '2_Spring', 
                       '2_Summer', '3_Summer', '3_Summer', '3_Summer', 
                       '4_Autumn', '4_Autumn', 
                       '4_Autumn', '1_Winter']
            month_to_season = dict(zip(range(1,13), seasons))
            seasons = dtindex.month.map(month_to_season).values

        if period=='half-year':
            seasons = ['Winter', 'Winter', 'Winter', 'Summer', 'Summer', 
                       'Summer', 'Summer', 'Summer', 'Summer', 'Winter', 
                       'Winter', 'Winter']
            month_to_season = dict(zip(range(1,13), seasons))
            seasons = dtindex.month.map(month_to_season).values

        return seasons


    def plot_time(self,figpath=None,figsize=None,colors=None):
        """Plot al heads in one graph and all head differences below that 
           in one figure """

        if figpath is None:
            msg = f'Parameter figpath must be given'
            raise ValueError(msg)

        if figsize is None:
            figsize = [12,18]

        if colors is None:
            colors = sns.color_palette('bright')

        nrows = len(list(self._heads))+1
        fig,axes = plt.subplots(nrows,1)
        plt.subplots_adjust(hspace=0.4)
        fig.set_figwidth(figsize[0])
        fig.set_figheight(figsize[1])

        for i,col in enumerate(list(self._heads)):

            title = f'{self.locname} (all series)'
            
            self._heads[col].dropna().plot(ax=axes[0],color=colors[i],
                                           title=title)
            sr = self._headsdif[col]*100
            sr.dropna().plot(ax=axes[i+1],color=colors[i])

            axes[i+1].set_title(sr.name, fontsize='large')
            axes[i+1].yaxis.set_label_text('stijghoogteverschil (cm)')

            axes[i+1].axhline(y=0, color='darkgray',linestyle='--')

        for i in range(len(axes)):
                axes[i].set_xlabel(' ')
        #fig.suptitle(locname,fontsize=12)

        fig.savefig(figpath, dpi=300, facecolor='w', edgecolor='w')

        return fig,axes


    def plot_head(self,figpath=None,color='season'):
        """Plot head difference by reference head value."""

        colnames = self._headsdif.columns

        if len(self._headsdif.columns) <= 2:
            ncols = 1
            nrows = 1
            figsize = (8.5,8.5)
        else:
            naxs = len(colnames)-1
            ncols = 2
            nrows = math.ceil(naxs/ncols)
            figsize = (8.5,4*nrows)

        fig, axs = plt.subplots(
                    nrows, ncols, squeeze=False, 
                    sharex=True, sharey=False,
                    figsize=figsize, dpi=100)

        axix = [(r,c) for r in range(nrows) for c in range(ncols)]

        if color=='half-year':
            seasons = self.date_seasons(self._headsdif,periods='half-year')
            coldict = {'Winter':'#0057e7','Zomer':'#ffa500'}
            colors = [coldict[x] for x in seasons]

        if color=='season':
            seasons = self.date_seasons(self._headsdif,period='quarter')
            coldict = {'1_Winter':'#0057e7', '2_Voorjaar':'#74d4f0',
                       '3_Zomer':'#ff0000','4_Herfst':'#ffa500'}
            colors = [coldict[x] for x in seasons]

        pal = sns.color_palette(colors)

        for i in range(len(colnames)-1):

            ax = axs[axix[i]]

            ax.axhline(y=0, color='darkgray',linestyle='--')
            ax.axvline(x=0, color='darkgray',linestyle='--')

            ax.axhline(y=5, color='slategray',linestyle='-.')
            ax.axhline(y=-5, color='slategray',linestyle='-.')

            data = self._headsdif
            data[self._refcol] = self._headsref[self._refcol]

            ax = sns.scatterplot(
                    x = colnames[0], 
                    y = colnames[i+1], 
                    data=data*100, 
                    ax=ax,
                    #color=colors[i+1])
                    palette=coldict,
                    hue = seasons)

            handles, labels = ax.get_legend_handles_labels()
            # sort both labels and handles by labels
            labels, handles = zip(*sorted(zip(labels, handles), key=lambda t: t[0]))
            ax.legend(handles, labels)

            ax.set_title(colnames[i+1])
            xlab = f'relatieve stijghoogte {colnames[0]} (cm)'
            ax.xaxis.set_label_text(xlab)
            ylab = 'stijghoogteverschil (cm)'
            ax.yaxis.set_label_text(ylab)

        fig.tight_layout()
        fig.savefig(figpath, dpi=300)

        return fig, axs


    def plot_freq(self):
        """
        Plot head differences as grid of frequency plots.

        Returns
        -------
        fig,ax
        """
        loclist = list(self._headsdif.columns)
        nrows = len(loclist)
        ncols = len(loclist)
        fig, axs = plt.subplots(nrows=nrows,ncols=ncols,sharex=True)

        # space between subplots
        plt.tight_layout()
        bbox = axs[0,0].get_position()
        hpad = bbox.width/100.*10.
        vpad = bbox.height/100.*10.
        plt.subplots_adjust(bottom=0.1,top=0.9,left=0.1,wspace=vpad, 
                            hspace=hpad)

        for i,ref in enumerate(loclist):

            tbl = self.table_headsdif(ref)
            for j,col in enumerate(loclist):
                x = (tbl[col]-tbl[col].mean()).values*100
                
                xmax=22.5
                dx = 5.

                nbars = math.ceil((xmax-dx/2)/dx)
                xup = nbars*dx+dx/2
                xlow = -xmax
                
                bin_list = np.arange(xlow,xup+1,dx)
                n, bins, patches = axs[i,j].hist(x, bin_list, density=1,)

                # color of rectangles (this is a slow solution)
                for k,pt in enumerate(patches):
                    if k < len(bin_list)/2-1:
                        pt.set_facecolor('#4ca3dd')
                    elif k == len(bin_list)/2-1:
                        pt.set_facecolor('#ccdee9') # #c0d6e4
                    else:
                        pt.set_facecolor('#ff7f50')

                axs[i,j].set_yticks([])
                axs[i,j].set_xticks([])

                if i==0:
                    axs[i,j].set_title(f'{col}',fontsize=18)
                if j==0:
                    axs[i,j].set_ylabel(f'{ref}',fontsize=18)

                # title below graphs
                if i ==nrows-1:
                    bbox = axs[i,j].get_position()
                    vpad = bbox.height/10.
                    mid = bbox.xmin+bbox.width/2.
                    plt.figtext(mid,bbox.ymin-vpad,f'{dx} cm',
                                fontsize='large',
                                horizontalalignment='center')

        return fig,axs



""" Read groundwater measurement data from dinoloket csv files.

  Author : Thomas de Meij, 2020
  
  History: 02-02-2014 created for python2.7;
           15-08-2015 migrated to python3.x;     
           06-07-2019 migrated to acequia 

"""

import os
import os.path
from datetime import datetime
import csv
import time
import datetime as dt
import warnings
import numpy as np
from pandas import Series, DataFrame
import pandas as pd

sep = ","

class DinoGws:
    """Read TNO Dinoloket csv file with groundwater measurement data"""

    _metatag = ','.join(
        ['Locatie','Filternummer','Externe aanduiding',
         'X-coordinaat','Y-coordinaat','Maaiveld (cm t.o.v. NAP)',
         'Datum maaiveld gemeten','Startdatum','Einddatum',
         'Meetpunt (cm t.o.v. NAP)','Meetpunt (cm t.o.v. MV)',
         'Bovenkant filter (cm t.o.v. NAP)',
         'Onderkant filter (cm t.o.v. NAP)'
        ])

    _datatag = ','.join(
        ['Locatie','Filternummer','Peildatum',
         'Stand (cm t.o.v. MP)','Stand (cm t.o.v. MV)',
         'Stand (cm t.o.v. NAP)','Bijzonderheid,Opmerking','','',''
         ])

    _missingdata = (f'Van deze put zijn geen standen opgenomen',
         f'in de DINO-database')

    _header_cols = ["nitgcode","filter","tnocode","xcoor",
        "ycoor","mvcmnap","mvdatum","startdatum","einddatum",
        "mpcmnap","mpcmmv","filtopcmnap","filbotcmnap"]

    _data_cols = ["nitgcode","filter","peildatum","standcmmp",
        "standcmmv","standcmnap","bijzonderheid","opmerking"]

    _head_cols = ['peildatum','standcmmp','bijzonderheid','opmerking']


    def __repr__(self):
        return (f'{self.srname()} (n={len(self._data)})')

    def __init__(self,filepath=None,readall=True):

        # lines marking data blocks in dinofiles

        if isinstance(readall,bool):
            self.readall=readall
        else:
            self.readall=True
            wrnstr = 'Variable \'{vname}\' not of type boolean' \
                     'of type \'{tname}\'. Data wil be read.'.format(
                     vname=readall,tname=type(readall))
            warnings.warn(warnstr)


        # herekenningsregels dinofiles
        self._metatag = "Locatie,Filternummer,Externe aanduiding,X-coordinaat,Y-coordinaat,Maaiveld (cm t.o.v. NAP),Datum maaiveld gemeten,Startdatum,Einddatum,Meetpunt (cm t.o.v. NAP),Meetpunt (cm t.o.v. MV),Bovenkant filter (cm t.o.v. NAP),Onderkant filter (cm t.o.v. NAP)"
        self._datatag = "Locatie,Filternummer,Peildatum,Stand (cm t.o.v. MP),Stand (cm t.o.v. MV),Stand (cm t.o.v. NAP),Bijzonderheid,Opmerking,,,"
        self._missingdata = "Van deze put zijn geen standen opgenomen in de DINO-database"
        self._header_cols = ["nitgcode","filter","tnocode","xcoor","ycoor","mvcmnap","mvdatum","startdatum","einddatum","mpcmnap","mpcmmv","filtopcmnap","filbotcmnap"]
        self._data_cols = ["nitgcode","filter","peildatum","standcmmp","standcmmv","standcmnap","bijzonderheid","opmerking"]

        # create empty variables
        self._reset()
        
        if filepath != None:
            self.flines = self._readfile(filepath)
            self._header, self._data = self._readlines()

        if self._header.empty and not self._data.empty:
            self._header = DataFrame(data=[[np.nan]*len(self._header_cols)],columns=self._header_cols)
            self._header = self._header.astype({'nitgcode':str,'filter':str})
            self._header.at[0,'nitgcode'] = self._data.at[0,'nitgcode']
            self._header.at[0,'filter'] = self._data.at[0,'filter']
            self._header.at[0,'startdatum'] = self._data.at[0,'peildatum']
            #self._tubeprops.at[0,'startdate'] = heads.index[0]


    def _reset(self):
        """ Reset all variables """
        self.filepath = ""
        self.errors = []
        self._data = DataFrame()
        self._header = DataFrame()       
        self._datatext = DataFrame()
        self._headertext = DataFrame()
        self.dfdesc = DataFrame()
        self.dfdescloc = DataFrame()
        self.seriesname = ""

    def _readfile(self,filepath):
        """ Open DINO csv file and return list of filelines """

        self._reset()
        self.filepath = filepath

        try:
            self.file = open(self.filepath,'r')
        except (IOError, TypeError) as err:
            errno, strerror = err.args
            print("{!s}".format(errno), end="")
            print("I/O fout{!s}".format(strerror), end="")
            print (" : "+self.filepath)
            self.errors.append(
                    [self.filepath,
                     "File can not be opened"])
            self.flines=[]
            raise
        else:
            self.flines = self.file.readlines()
            self.file.close()

        return self.flines


    def _readlines(self):
        """ read list of file lines from dinofile to data """

        # assert file is valid ascii
        if len(self.flines)==0:
            file_valid = False
        elif self.flines[0][0]=='\x00':
            # test for corrupted file with only 'x00' 
            # Yes, I have really seen this
            file_valid=False
        else:
            file_valid = True

        if file_valid==False:
            self.headerstart=0
            self.headerend=0
            self.datastart=0
        else:
            # findlines in dinofile lines
            self.headerstart, self.headerend, self.datastart = \
                self._findlines()
        
        # read header
        if self.headerstart>0 and self.headerend>0: 
            self._header = self._readheader()
        else:
            self._header = DataFrame()

        # read data
        if self.datastart>0:
            self._data = self._readgws()
        else:
            self._data = DataFrame()

        return self._header, self._data


    def _findlines(self):
        """ Find start of header and data; if file has data at all """

        # set variables to find at zero
        self.headerstart = 0
        self.headerend = 0
        self.datastart = 0

        # find variables
        for il in range(len(self.flines)):

            if self.flines[il].startswith(self._missingdata):
                # put zonder gegevens
                self.errors.append([self.filepath,"Bestand bevat geen data"])
                self.hasheader = False
                self.hasdata = False
                break

            if self.flines[il].startswith(self._metatag): #("Locatie,Filternummer,Externe"):
                if not self.flines[il+1].startswith("B"): # er zijn geen headerlines onder de headerkop
                    self.hasheader = False
                    self.errors.append([self.filepath,"Bestand zonder header"])                               
                else:
                    while True:
                        il+=1
                        if self.flines[il].startswith("B"):
                            if self.headerstart==0:
                                self.hasheader = True                        
                                self.headerstart = il
                                #self.headerlength = 1
                            #else:
                            #    self.headerlength+=1
                        else: #voorbij laatste regel header
                            self.headerend = il
                            break
                            
            if self.flines[il].startswith(self._datatag): #("Locatie,Filternummer,Peildatum"):
                # bepaal eerste regelnummer met data
                il+=1
                if self.flines[il].startswith("B"):
                    self.hasdata = True
                    self.datastart = il
                else:
                    self.hasdata = False
                    self.errors.append([self.filepath,"Bestand zonder grondwaterstanden"])
                break
        il+=1
        # end of def findlines
        return self.headerstart, self.headerend, self.datastart

    @staticmethod
    def parse_dino_date(datestring,addtime=False):
        if isinstance(datestring, str):
            if datestring!="":                
                # string to datetime.datetime object
                if addtime==True: date = datetime.strptime(
                    datestring+" 12:00", "%d-%m-%Y %H:%M")
                else: date = datetime.strptime(datestring, "%d-%m-%Y")                    
                
                ## replace invalid date with np.NaN
                ##if date.year < 1900: date = np.NaN
                ##elif date.year > datetime.now().year: date = np.NaN

            else:
                date = np.NaN
        else:
            date = np.NaN
        return date

    def _readheader(self): #public
        """ Read header data and return pandas dataframe """


        if self.headerstart>0 and self.headerend > self.headerstart:
            # create _header
            headerlist = [line[:-1].split(sep) for line in self.flines[self.headerstart:self.headerend]]
            self._header = DataFrame(headerlist, columns=self._header_cols)
            self._headertext = DataFrame(headerlist, columns=self._header_cols)

            # transform column values
            self._header["filter"] = self._header["filter"].apply(lambda x: x.lstrip("0"))
            self._header["mvdatum"] = self._header["mvdatum"].apply(lambda x:self.parse_dino_date(x))
            self._header["startdatum"] = self._header["startdatum"].apply(lambda x:self.parse_dino_date(x))
            self._header["einddatum"] = self._header["einddatum"].apply(lambda x:self.parse_dino_date(x))

            # make seriesname
            ##self.seriesname = self._header["nitgcode"].values[0]+"_"+str(self._header["filter"].values[0])
        else:
            # create empty dataframe
            self._header = DataFrame(columns=self._header_cols)
            self._headertext = DataFrame(columns=self._header_cols)
            self.seriesname = "onbekend" # self.filename.split(".")[0]
            #print("warning : series has no header")
        return self._header


    def _readgws(self):
        """ Read groundwater measurements to pandas data frame """

        def fstr2float(astr):
            try:
                aval = float(astr)
            except ValueError:
                aval = np.NaN
            return aval


        if self.datastart>0:

            # create list of data from filelines
            data = [line[:-1].split(sep)[0:7]+[sep.join(line[:-1].split(sep)[7:])] for line in self.flines[self.datastart:]]            
            self._datatext = DataFrame(data,columns=self._data_cols)
            self._data = DataFrame(data,columns=self._data_cols)

            # transform column values
            self._data["peildatum"] = self._data["peildatum"].apply(lambda x:datetime.strptime(x,"%d-%m-%Y"))  # 28-03-1958
            self._data["filter"] = self._data["filter"].apply(lambda x: x.lstrip("0"))
            self._data["standcmmp"] = self._data["standcmmp"].apply(lambda x: fstr2float(x))
            self._data["standcmmv"] = self._data["standcmmv"].apply(lambda x: fstr2float(x))
            self._data["standcmnap"] = self._data["standcmnap"].apply(lambda x: fstr2float(x))
            self._data["opmerking"] = self._data["opmerking"].apply(lambda x: x.strip(","))

        else:
            self._data = DataFrame(columns=self._data_cols)
            self._datatext = DataFrame(columns=self._data_cols)
        return self._data

    def series(self,units="cmmv"):
        """ Return groundwater measurements as pandas time series 
        Possible values for parameter units are : "cmmv", "cmmp" or "cmnap"
        """
        # create series from _data
        if len(self._data)>0:
            if units=="cmmv":
                # data = self.data[self.data["standcmmv"]!=""] # alle ontbrekende waarden eruit
                self.srseries = Series(self._data["standcmmv"].values, index=self._data["peildatum"],name=self.srname())
            if units=="cmmp":
                self.srseries = Series(self._data["standcmmp"].values, index=self._data["peildatum"],name=self.srname())
            if units=="cmnap":
                self.srseries = Series(self._data["standcmnap"].values, index=self._data["peildatum"],name=self.srname())
        else: # create empty series
            self.srseries = Series(name=self.srname())
        return self.srseries


    def headdata(self):
        """Return head data fromdino csv file"""

        if len(self._data)>0:
            data = self._data[self._head_cols].copy()
        else:
            data = DataFrame(columns=self._head_cols)

        return data


    def header(self):
        """ Return raw metadata from series as a pandas dataframe """        
        return self._header


    def data(self):
        """ Return raw data as Pandas dataframe """
        return self._data


    def locname(self,newname=None):
        if newname!=None: 
            self.location = newname
        elif not self.header().empty:
            self.location = self._header.loc[0,"nitgcode"]
            ##self.location = self._header.at[0,"nitgcode"]
        else:
            self.location = "B00A0000"
        return self.location


    def filname(self,newname=None):
        if newname!=None: 
            self.filter = newname
        elif not self.header().empty:
            self.filter = self.header().loc[0,"filter"]
        else: self.filter = "0"
        return self.filter


    def srname(self):
        self.seriesname = self.locname()+"_"+self.filname()
        return self.seriesname


    def describe(self):
        """ Return one line of metadata from series as a pandas dataframe """

        # deze functie moet eruit geschreven worden omdat parse_dino_date dit al afvangt
        def valdate(date):
            """ Validate datetime (no dates before 1900) """
            return (date.strftime("%d-%m-%Y") if date.year>1900 else np.NaN)

        if self.dfdesc.empty:
            filcols = ['reeksnaam', 'nitgcode', 'filter', 'tnocode', 'xcoor', 'ycoor','mvcmnap', 'mvdatum', 'startdatum', 'einddatum', 'mpcmnap', 'mpcmmv','filtopcmnap', 'filbotcmnap']
            self.dfdesc = DataFrame(columns=filcols)

            if len(self._readheader()) !=0 and len(self._readgws())!=0:

                # make one line of metadata
                dftail = self.header().tail(1).copy()

                # insert series name as first column
                ##dftail['reeksnaam'] = dftail["nitgcode"]+"_"+dftail["filter"].values.astype('str')
                dftail['reeksnaam'] = self.srname()

                # recalculate startdate and enddate from measurements
                colnr = dftail.columns.get_loc("startdatum")
                startdatestr = self.data()["peildatum"][0]
                dftail.iloc[0,colnr] = pd.to_datetime(valdate(startdatestr),format='%d-%m-%Y')
                
                colnr = dftail.columns.get_loc("einddatum")
                enddatestr   = self.data()["peildatum"][len(self.data())-1]
                dftail.iloc[0,colnr]  = pd.to_datetime(valdate(enddatestr),format='%d-%m-%Y')

                self.dfdesc = dftail.reindex(columns=filcols).copy()
                self.dfdesc = self.dfdesc.reindex() # set index to 0

        return self.dfdesc


    def mpref(self):
        """ create dataframe with series of mp reference changes (for plotting line of ref changes above gwseries graph)"""
        msg = 'mpref method is depricates. use GwSeries.tubepropchanges() instead.'
        warnings.warn(msg, warnings.DeprecationWarning)


    def locations(self,df=DataFrame()):
        """ create table of locations from dataframe with data from several filters created by function describe() """

        # select one row for each group of filters
        grp = df.groupby("nitgcode")
        dfloc = grp.last()
        dfloc["startdatum"] = grp["startdatum"].min()
        dfloc["einddatum"] = grp["einddatum"].max()
        dfloc["nfil"] = pd.to_numeric(grp["filter"].count())

        # add index as column and reindex rows
        dfloc.insert(0,"nitgcode",dfloc.index)
        dfloc = dfloc.reset_index(drop=True)

        # drop columns with filter specific data
        dropcols = ["reeksnaam","filter",'mpcmnap','mpcmmv','filtopcmnap','filbotcmnap']
        dfloc = dfloc.drop(dropcols,axis=1,inplace=False) #,axis=1)

        # reorder columns
        dfloc = dfloc.reindex(columns=['nitgcode','tnocode','nfil','mvcmnap','mvdatum','startdatum','einddatum','xcoor','ycoor'])

        return dfloc


    def merge(self,dn2):
        """ add series of dinogws object to dinogws object and return merged object """

        if not isinstance(self,dinogws):
            raise TypeError("Input of dinogws.merge() must be another dinogws object")

        
        if (not self._data.empty) and (not dn2._data.empty):

            # determine possible overlap
            startdate1 = self.data().peildatum.min()
            enddate1 = self.data().peildatum.max()
            startdate2 = dn2.data().peildatum.min()
            enddate2 =  dn2.data().peildatum.max()
            if (startdate2 <= enddate1) and (startdate1 <= enddate2):
                raise ValueError('date ranges from both series show overlap')

            # merge headers
            self._header = pd.concat([self._header,dn2._header])
            self._header = self._header.sort_values(by=['startdatum']).reset_index()

            # merge groundwater measurement data
            self._data = pd.concat([self._data,dn2._data]).sort_values(by=['peildatum'])

        return self


def filesfromdir(dir):
    """Return list of dino sourcefiles from directory """

    filenames = []
    seriesnames = []
    for root, dirs, files in os.walk(sourcedir):
        seriesnames += [f[0:11] for f in files if f[11:13]=="_1"]
        filenames += [os.path.join(root, f) for f in files if f[11:13]=="_1"]
    return filenames, seriesnames




""" This module contains the KnmiLocations class that returns a list of
all available KNMI stations on the KNMI website and creates tables with
KNMI stations numbers and names

Example
-------
>>> aq.knmilocations()

"""

import os,os.path
import requests
import pkgutil
from io import StringIO
import warnings
import numpy as np
import pandas as pd
from pandas import Series, DataFrame
import pandas as pd

import acequia as aq
from acequia import CrdCon


def knmilocations(stntype='prc'):
    """ Return table of KNMI stations as pd.DataFrame 
    
    Parameters
    ----------
    stntype : 'prc','wtr','all'
        type of knmi station

    Returns
    -------
    pd.DataFrame

    """

    knmi = aq.KnmiStations()

    if stntype not in ['prc','wtr','all']:
        msg = f'Variable stntype must be "prc","wtr" or "all" '
        raise ValueError(smg)
    
    if stntype=='prc':
        tbl = knmi.prc_stns()

    if stntype=='wtr':
        tbl = knmi.wtr_stns()

    if stntype=='all':
        tbl1 = knmi.prc_stns()
        tbl2 = knmi.wtr_stns()
        tbl = pd.concat([tbl1,tbl2])

    xnull = pd.isnull(tbl.xcr)
    ynull = pd.isnull(tbl.ycr)
    crnan = list(tbl[(xnull|ynull)].stn_name.values)
    if len(crnan)!=0:
        msg = [f'{len(crnan)} stations with missing coordinates removed ',
               f'from tabel: {crnan}']
        warnings.warn(''.join(msg))

    return tbl[(~xnull & ~ynull)]


class KnmiStations:
    """Retrieve KNMI list of all available station numbers and names
    from KNMI website 

    Methods
    -------
    prc_stns(filepath=None)
        return list of manual rain gauche locations

    wtr_stns(filepath=None)
        return list of weather station locations

    Note
    ----
    Downloading lists of KNMI stations from the website takes time. 
    Recommended use is to download these data only once and save 
    results as csv for further use.

    """

    PRC_CRD_FILE = 'precipitation_crd.csv'
    STN_COLS = ['stn_name','stn_type','xcr','ycr','lon','lat','alt']
    STN_INDEX = 'stn_nr'


    def __init__(self):

        # read list of precipitation station coordinates from 
        # local csv file within package
        try:
            csvfile = pkgutil.get_data(__package__,self.PRC_CRD_FILE)
            data = StringIO(str(csvfile,'latin-1'))
            self._prc_crd = pd.read_csv(data, index_col='stn_nr')       
        except FileNotFoundError:
            self._prc_crd = DataFrame()
            msg = [f'Local file {self.PRC_CRD_FILE} not found. ',
                   f'Coordinates in list of precipitation stations ',
                   f'will all be NaN.']
            warnings.warn(''.join(msg))


    def prc_stns(self,filepath=None):
        """ Return table of all available precipitation stations on KNMI site
        
        Parameters
        ---------
        filepath : str, optional
            filepath for writing csv file with precipitation station list

        Returns
        -------
        pd.DataFrame

        Example
        -------
        knmi = KnmiStations()
        dfprc = knmi.prc_stns('prc_stations.csv')

        Note
        ----
        Coordinates of precipitation stations are not available on the
        KNMI website. This functions reads coordinates from a local csv 
        file.

        Explanation on KNMI website can be found on:
        https://www.knmi.nl/kennis-en-datacentrum/achtergrond/data-ophalen-vanuit-een-script

        """

        par = {}
        str1 = r'http://projects.knmi.nl/klimatologie/monv' 
        str2 = r'/reeksen/getdata_rr.cgi'
        url = str1+str2
        self._flines = requests.get(url,params=par).text.splitlines()

        prc_dict={}
        startline=-1
        for i,line in enumerate(self._flines):

            if startline>-1: # and i>=startline:
                linelist = line.split(',')
                stn = int(linelist[0])
                stn_name = linelist[4].strip()
                if not stn in prc_dict.keys(): 
                    prc_dict[stn]={'stn_name':stn_name}

            if line.startswith('STN,YYYYMMDD,'):
                startline=i+1

        prcstn = pd.DataFrame.from_dict(prc_dict, orient='index')
        #prcstn.index = prcstn.index.astype(dtype='str')
        prcstn.index.name = self.STN_INDEX
        prcstn['stn_type'] = 'prc' 

        nancols = [x for x in self.STN_COLS if x not in list(prcstn)]
        for col in nancols:
            prcstn[col]=np.nan
        newcols = [x for x in list(prcstn) if x not in self.STN_COLS]
        cols = self.STN_COLS+newcols
        prcstn = prcstn[cols]

        # fill in missing coordinates
        if not self._prc_crd.empty:
            xcrdict = self._prc_crd['xcrd'].to_dict()
            ycrdict = self._prc_crd['ycrd'].to_dict()
            for index,row in prcstn.iterrows():
                if index in xcrdict.keys():
                    prcstn.at[index,'xcr']=xcrdict[index]
                if index in ycrdict.keys():
                    prcstn.at[index,'ycr']=ycrdict[index]

            crdcon = CrdCon()
            for stn,sr in prcstn.iterrows():
                crdict = crdcon.RDtoWGS84(sr['xcr'],sr['ycr'])
                prcstn.at[stn,'lon'] = round(crdict['Lon'],3)
                prcstn.at[stn,'lat'] = round(crdict['Lat'],3)

        nmiss_xcr = prcstn['xcr'].isnull().sum()
        nmiss_ycr = prcstn['ycr'].isnull().sum()
        nmiss = max([nmiss_xcr,nmiss_ycr])
        if nmiss!=0:
            msg = [f'List of precipitation stations has {nmiss} ',
                   f'missing coordinates in total of {len(prcstn)} ',
                   f'stations.']
            warnings.warn(''.join(msg))

        if filepath:
            prcstn.to_csv(filepath,index=True)

        return prcstn


    def wtr_stns(self,filepath=None):
        """ Return table of all available KNMI weather stations 

        Parameters
        ---------        
        filepath : str, optional
            filepath for outputfile with list of wheater stations

        Returns
        -------
        pd.DataFrame

        Example
        -------
        knmi = KnmiStations()
        dfprc = knmi.wtr_stns('wtr_stations.csv')

        Note
        ----
        Explanation on KNMI website can be found on:
        https://www.knmi.nl/kennis-en-datacentrum/achtergrond/data-ophalen-vanuit-een-script

        """

        HEADERLINE = '# STN      LON(east)   LAT(north)     ALT(m)  NAME'
        DATALINE = '# YYYYMMDD'

        # daily data from meteorological stations
        str1 = r'http://projects.knmi.nl/klimatologie'
        str2 = r'/daggegevens/getdata_dag.cgi'
        url=str1+str2
        par = {'stns':'ALL'}
        self._flines = requests.get(url,params=par).text.splitlines()

        # find first and last line of table with data in text
        startline=-1    
        for i,line in enumerate(self._flines):

            if line.startswith(HEADERLINE):
                startline=i+1

            if startline!=-1 and line.startswith(DATALINE):
                endline = i-2
                break

        # table stn numbers and metadata
        wht_dict={}
        for i,line in enumerate(self._flines[startline:endline+1]):

            linelist = line.split()
            stn = linelist[1].rstrip(':')
            stn_name = ' '.join(linelist[5:])
            lon = linelist[2]
            lat = linelist[3]
            alt = linelist[4]

            if not stn in wht_dict.keys(): 
                wht_dict[stn]={
                    'lon':lon,
                    'lat':lat,
                    'alt':alt,
                    'stn_name':stn_name,
                    }

        wtrstn = pd.DataFrame.from_dict(wht_dict, orient='index')
        wtrstn.index.name = self.STN_INDEX
        wtrstn['stn_type'] = 'wtr'

        nancols = [x for x in self.STN_COLS if x not in list(wtrstn)]
        for col in nancols:
            wtrstn[col]=np.nan

        newcols = [x for x in list(wtrstn) if x not in self.STN_COLS]
        cols = self.STN_COLS+newcols
        wtrstn = wtrstn[cols]

        crdcon = CrdCon()
        for stn,sr in wtrstn.iterrows():
            crddict = crdcon.WGS84toRD(sr['lon'],sr['lat'])
            wtrstn.at[stn,'xcr'] = round(crddict['xRD'],0)
            wtrstn.at[stn,'ycr'] = round(crddict['yRD'],0)

        nmiss_xcr = wtrstn['xcr'].isnull().sum()
        nmiss_ycr = wtrstn['ycr'].isnull().sum()
        nmiss = max([nmiss_xcr,nmiss_ycr])
        if nmiss!=0:
            msg = [f'List of precipitation stations has {nmiss} ',
                   f'missing coordinates in total of {len(wtrstn)} ',
                   f'stations.',]
            warnings.warn(''.join(msg))

        if filepath:
            wtrstn.to_csv(filepath,index=True)

        return wtrstn





from pathlib import Path
import warnings
import numpy as np
import pandas as pd
from pandas import DataFrame, Series


class KnmiRain:
    """Read precipitation data from knmi file"""

    NAMESTR = 'STN,YYYYMMDD,   RD,   SX,' + 'empty'
    ##COLNAMES = ['STN','YYYYMMDD','RD','SX']
    VARNAMES = ['prc','snow']
    SNOWCODE = {
        '997': 'gebroken sneeuwdek/broken snow cover < 1 cm',
        '998': 'gebroken sneeuwdek/broken snow cover >=1 cm',
        '999': 'sneeuwhopen/snow dunes}',
        }
    SNOWVALS = {'997':'1','998':'1','999':'1'}
    SKIPROWS = 24

    def __init__(self,fpath):

        self.colnames = self.NAMESTR.replace(' ','').split(',')
        # NAMESTR[:-1] : trailing colon creates extra column
        self.keepcols = self.colnames
        self.fpath = Path(fpath)

        if not (self.fpath.exists() and self.fpath.is_file()):
            msg = [f'Filepath \'{self.fpath}\' not found',
                'empty dataframe is returned.']
            warnings.warn(' '.join(msg))
            self.rawdata = pd.DataFrame(columns=self.colnames)
            self.data = pd.DataFrame(columns=self.keepcols)
        else:
            self._readfile()
            self._clean_rawdata()


    def _readfile(self):

        # read csv to pd.DataFrame with only str values
        self.rawdata = pd.read_csv(self.fpath,sep=',',
            skiprows=self.SKIPROWS,
            names=self.colnames,dtype='str')

        # replace empty strings with NaN
        self.rawdata = self.rawdata.apply(
            lambda x: x.str.strip()).replace('', np.nan)


    def _clean_rawdata(self):

        self.data = self.rawdata[self.keepcols].copy()
        for colname in list(self.data):

            if colname=='YYYYMMDD':
                self.data[colname] = pd.to_datetime(self.data[colname],
                    infer_datetime_format=True)
                self.data = self.data.set_index(
                    colname,verify_integrity=True)
                self.data.index.name='date'

            if colname=='RD':
                self.data[colname] = self.data[colname].astype(float)/10.

            if colname == 'SX':
                self.data[colname] = self.data[colname].replace(self.SNOWVALS)
                self.data[colname] = self.data[colname].astype(float)

            if colname == 'empty':
                self.data = self.data.drop(columns=[colname])


    def timeseries(self,varname):
        """Return timeseries with data

        Parameters
        ----------
        varname : {'prc')
            variable to return
        """
 
        if varname not in self.VARNAMES:
            msg = [f'{varname} is not a valid variable name.',
                f'parameter varname must be in {self.VARNAMES}',
                f'by default rain data are returned']
            warnings.warn(' '.join(msg))
            varname = 'prc'

        if varname == 'prc':
            sr = self.data['RD']
            sr.name = 'prc'

        first = sr.first_valid_index()
        last = sr.sort_index(ascending=False).first_valid_index()
        return sr[first:last]


    def units(self):
        """Return table with definitions and units of variables"""

        tbl = pd.DataFrame({
            'variable' : ['prc','snow'],
            'datacol' : ['RH','EV24'],
            'unit' : ['mm/day','cm/day'], 
            })
        tbl = tbl.set_index('variable')
        return tbl






""" This module contains the object GwLocs


"""

import os
import warnings
from pandas import Series, DataFrame
import pandas as pd
import numpy as np

import acequia as aq


class GwLocs:
    """Manage multiple groundwater heads series from one well location

    Parameters
    ----------
    filedir : str
        directory with source files

    pathlist : list, optional
        list of sourcefile names (without path)

    filetype : ['.csv','.json'], optional
        source file type 

    groups : list, optional
        list of location names, sublists are allowed (see examples)

    Methods
    -------
    sourcetable()
        return list of location and heads series names

    gwseries(self,loc=None)
        return list of GwSeries objects

    Examples
    --------
    Create GwLocs object:

    >>>locs = GwLocs(filedir=<jsondir>)

    Return table of locations and series:

    >>>tbl = locs.loctable()

    Return al series for location B16D0037:

    >>>gws = GwLocs.gwseries(loc='B16D0037')

    Return list of GwSeries objects for locations in list:

    >>>names = ['B16D0037','B27G0237',['B28D1635','B28B1388'],'B28B1389']

    >>>for loc in GwLocs(filedir=<jsondir>,groups=names):

            print(f'{names[i]} group size is {len(gws)}')

    Explicitly iterate over locations:

    >>>locs = GwLocs(filedir=<jsondir>,groups=names)

    >>>for i in range(len(locs)):

            gws = next(locs)

            print(f'{names[i]} group size is {len(gws)}')

    """

    def __init__(self,filedir=None,pathlist=None,filetype=None,
        groups=None):
        """Return GwLocs object"""

        self._filedir = filedir
        self._pathlist = pathlist
        self._filetype = filetype
        self._loclist = groups

        if not isinstance(self._filedir,str):
            msg = f'filedir must be a string, not {self._filedir}'
            raise ValueError(msg)

        if not os.path.isdir(self._filedir):
            msg = f'{self._filedir} is not a valid directory'
            raise ValueError(msg)

        if self._pathlist is None:
            self._pathlist = aq.listdir(dirpath=self._filedir)

        if self._filetype is None:
            self._filetype = self._infer_filetype()

        if self._filetype in ['json','csv']:
            self._filetype = '.'+self._filetype

        if self._filetype not in ['.json','.csv']:
            msg = f'Unsupported filetype {filetype}.'
            raise ValueError(msg)

        ##self._srcfiles = self._srctbl()
        ##self._locfiles = self.list_locfiles()
        self._create_iterator()


    def __repr__(self):
        nlocs = len(self._srctable().groupby(by='loc').size())
        return (f'GwLocs(nlocs={nlocs})')


    def _infer_filetype(self):
        """Infer filetype from file extensin count in <filedir>"""
        filelist = [os.path.basename(path) for path in self._pathlist]
        extlist = [os.path.splitext(filename)[1] for filename in filelist]
        sr = Series(extlist).value_counts()
        return sr.index[0]


    def _srctable(self):
        """Return table of all valid source files"""

        if self._filetype=='.json':
            filepaths = [x for x in self._pathlist]
            fnames = [os.path.basename(path) for path in self._pathlist]
            ftypes = [os.path.splitext(name)[1] for name in fnames]
            series = [os.path.splitext(name)[0] for name in fnames]
            locations = [x.split('_')[0] for x in series]
            filters = [x.split('_')[1] for x in series]

        if self._filetype=='.csv':
            # dino zipfiles come with two files for each filter
            # only files ending with _1.csv are needed
            # filename: B01C0008001_1.csv
            filepaths = [x for x in self._pathlist if x.endswith('_1.csv')]
            fnames = [os.path.basename(x) for x in filepaths]
            ftypes = [os.path.splitext(name)[1] for name in fnames]
            locations = [x[:8] for x in fnames]
            filters = [x[8:11].lstrip('0') for x in fnames]
            series = [x[0]+'_'+x[1] for x in list(zip(locations,filters))]

        table = DataFrame(data={
                'loc':locations,
                'fil':filters,
                'filename':fnames,
                'filetype':ftypes,
                'filepaths':filepaths,
                },index=series)
        table.index.name = 'series'
        return table


    """
    def list_sourcefiles(self,filedir=None):
        """ """Return table with list of series names and well location 
        names based on sourcefile names in directory filedir

        Parameters
        ----------
        filedir : str, optional
            directory with sourcefile names

        Returns
        -------
        pd.DataFrame

        Notes
        -----
        Directory <filedir> should contain source files with filenames
        composed of <location name>_<filter>.<extension>, like in 
        'B28H0025_1.csv' or 'Mylocation_Myfilter.json'.

        """ """

        if filedir is None:
            return self._srcfiles

        if not os.path.isdir(filedir):
            msg = f'{filedir} must be a valid directory'
            raise ValueError(msg)

        return self._srctable(filedir)
    """

    def filetable(self):
        """Return table of source files for all series"""

        srctbl = self._srctable()

        if self._loclist is not None:

            def flatten_nested_list(nested_list):
                # stackoverflow.com/questions/10823877/
                for i in nested_list:
                    if isinstance(i, (list,tuple)):
                        for j in flatten_nested_list(i):
                            yield j
                    else:
                        yield i

            flatlist = flatten_nested_list(self._loclist)
            mask = srctbl['loc'].isin(flatlist)
            srctbl = srctbl[mask]

        if srctbl.empty:
            msg = f'list_locfiles returns empty dataframe.'
            warnings.warn(msg)

        return srctbl


    def _create_iterator(self):
        """Create location series groups iterator """
        self._itercount = 0
        if self._loclist is None:
            ##grp = self._srcfiles.groupby(by='loc')
            ##grp = self._locfiles.groupby(by='loc')
            grp = self.filetable().groupby(by='loc')
            self._loclist = list(grp.groups.keys())
        else:
            self._loclist = self._loclist
            #todo: check if groupnames are in table


    def gwseries(self,loc=None):
        """Return list of GwSeries objects with location name <loc>

        Parameters
        ----------
        loc : str
            name of location

        Return
        ------
        list of GwSeries objects

        Notes
        -----
        When parameter loc is a list of locations names, GwSeries objects
        for all these locations will be returned.
        """

        if isinstance(loc,str):
            loc = [loc]

        mask = self._srctable()['loc'].isin(loc)
        gwlist = []
        for i,(srname,row) in enumerate(self._srctable()[mask].iterrows()):

            filepath = os.path.join(self._filedir,row['filename'])
            if self._filetype=='.json':
                gw = aq.GwSeries.from_json(filepath)
            if self._filetype=='.csv':
                gw = aq.GwSeries.from_dinogws(filepath)

            #if i==0:
            #    gwlist = []
            gwlist.append(gw)

        return gwlist


    def __iter__(self):
        return self


    def __len__(self):
        return len(self._loclist)


    def __next__(self):

        if self._itercount >= len(self):
            raise StopIteration

        loc = self._loclist[self._itercount]
        mygwlist = self.gwseries(loc=loc)

        self._itercount+=1
        return mygwlist

